% COMPUTE_BEZIER_SPLINE_PARAMS - compute missing n-dimensional cubic spline
%   control points from nodes.
%
% Parameters:
%    v - row vector of n-dimensional nodes of a closed curve which should
%        be interpolated
%
% Returns:
%    a, b - row vectors of the missing Bezier spline control points
%
% File        : compute_bezier_spline_params.m
% Date        : 2009-03-09
% Author      : Oliver Pajonk

% Rollercoaster Prototype - A MATLAB rollercoaster
% Copyright (C) 2009  Oliver Pajonk
% 
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
% 
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% 
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.

function [ a, b ] = compute_bezier_spline_params(v)
    % determine number of nodes and dimensions
    [dims nodes] = size(v);

    % Fill linear system matrix A
    % In out setup the matrix has four parts, structured in the following
    % way:
    % [ 1 | 2 ]
    % [ 3 | 4 ]
    % Why the matrix looks like this is not described in the SEP document -
    % but you can easily derive it yourself if you look at the system of
    % linear equations you get for the spline parameters!

    A = zeros(nodes*dims*2);
    for i=1:nodes
        for k=1:dims
            % Part 1
            A((i-1)*dims+k,(i-1)*dims+k) = 1/2; % 1/2 on the upper main diagonal

            % Part 2
            if (i ~= 1)
                A((i-1)*dims+k, nodes*dims + (i-2)*dims+k) = 1/2; % 1/2 on the upper diagonal
            end

            % Part 3
            A(nodes*dims +(i-1)*dims+k, (i-1)*dims+k) = 2; % 2 on the lower diagonal
            if (i ~= 1)
                A(nodes*dims + (i-1)*dims+k, (i-2)*dims+k) = 1; % 1 on the lower main diagonal
            end

            % Part 4
            A(nodes*dims +(i-1)*dims+k,nodes*dims + (i-1)*dims+k) = -1; % -1 on the lower main diagonal        
            if (i ~= 1)
                A(nodes*dims +(i-1)*dims+k,nodes*dims + (i-2)*dims+k) = -2; % -2 on the lower main diagonal
            end
        end
    end

    % Special cases of linear system matrix
    for k=1:dims
        % Part 2
        A(k, nodes*dims + (nodes-1)*dims+k) = 1/2;
        % Part 3
        A(nodes*dims + k, (nodes-1)*dims + k) = 1;
        % Part 4
        A(nodes*dims + k, (nodes-1)*dims + nodes*dims + k) = -2;
    end


    % Setup RHS
    rhs = zeros(1, nodes*dims*2);
    % We put the nodes into the first half of the RHS, simply concatenating
    % the nodes
    rhs(1:nodes*dims) = reshape(v,1,[]);

    % Solve system (this is very easy in MATLAB)
    sp = A \ rhs';

    % Prepare output of Bezier spline control points
    sp = reshape(sp, dims, nodes*2);
    a = sp(:,1:nodes);
    b = sp(:,nodes+1:2*nodes);
end