% RK4 - Runge-Kutta 4th order method to integrate ODEs
%   The interface is compatible with the MATLAB ODE solvers.
%
% Parameters:
%           ode - handle to the ODE to integrate
%             T - time points at which the ODE should be evaluated
%            s0 - start value(s)
%
% Returns:
%             T - time points at which the ODE was be evaluated (a copy of
%                 the input T, just to comply to the interface)
%             s - vector of solutions at the time points
%
% File        : rk4.m
% Date        : 2009-03-18
% Author      : Oliver Pajonk

% Rollercoaster Prototype - A MATLAB rollercoaster
% Copyright (C) 2009  Oliver Pajonk
% 
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
% 
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% 
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.

function [ T, s ] = rk4(ode, T, s0)
    s = zeros(2, length(s0));
    step = 1;
    s(step,:) = s0;
    while (step < length(T))
        step = step + 1;
        delt = T(step) - T(step-1);
        s(step,:) = runge_kutta_order4_step(ode, s(step-1,:)', delt);
    end
end

function [ snext ] = runge_kutta_order4_step(ode, s, delt)
    k1 = ode(delt, s);
    k2 = ode(delt, s + 0.5*delt.*k1);
    k3 = ode(delt, s + 0.5*delt.*k2);
    k4 = ode(delt, s + delt.*k3);
    snext = s + (delt/6)*(k1 + 2*k2 + 2*k3 + k4);
end