% ROLLERCOASTER - main file of the prototype solution for the
%   software development practical course at the Institute of Scientific
%   Computing, TU Braunschweig.
%
% File        : rollercoaster.m
% Date        : 2009-03-09
% Author      : Oliver Pajonk

% Rollercoaster Prototype - A MATLAB rollercoaster
% Copyright (C) 2009  Oliver Pajonk
% 
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
% 
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% 
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.

function rollercoaster()
    clear variables;

    % =====================================================================
    % Step 1: read in the track data
    %   In this prototype this is not in an XML format, only lists of
    %   nodes. The assumed unit for the data is meters
%     [x y z] = track_small_circle();
%     [x y z] = track_large_circle();
%     [x y z] = track_trapez_circle();
%     [x y z] = track_test_track();
    [x y z] = track_complex();

    % create shorthand for 3D-vectors of the nodes
    v = [x; y; z];
    % compute dimensions (this we need later at several places)
    [dims nodes] = size(v);

    % =====================================================================
    % Step 2: Compute the missing order three Bezier spline control points 
    %         from the nodes
    [a b] = compute_bezier_spline_params(v);
    
    % Compute the length of the rollercoaster track
    % This amounts to numerically calculating the arc length integrals for
    % each piece of the spline curve and summing it up
    % (not necessary, but handy)
    partial_lengths = zeros(1,nodes);
    for i = 1:nodes-1
        F = @(s)ds_bezier_curve(s, v(:,i), a(:,i), b(:,i), v(:,i+1));
        partial_lengths(i) = quad(F, 0, 1);
    end
    F = @(s)ds_bezier_curve(s, v(:,nodes), a(:,nodes), b(:,nodes), v(:,1));
    partial_lengths(i) = quad(F, 0, 1);
    track_length = sum(partial_lengths);
    fprintf(1, 'Length of the track: %f meters\n', track_length);

    % =====================================================================
    % Step 3: compute start values and solve the ODE    
    
    % Start position (in spline parameter space - an absolute position in
    % 3D space would not be very handy here...)
    s0 = 0;
    
    % Initial speed in 3D space (along the spline, so it is a scalar value,
    % not a vector)
    xdot0 = 33; % unit: meters/s
    
    % Compute sdot0 (initial speed in parameter space)
    %  first determine on which node s0 is and compute the geometry for
    %  it...
    node = mod(floor(s0), nodes) + 1;
    if (node ~= nodes)
        [x, xs] = bezier_curve(s0 - floor(s0), v(:,node), a(:,node), b(:,node), v(:,node+1));
    else
        [x, xs] = bezier_curve(s0 - floor(s0), v(:,node), a(:,node), b(:,node), v(:,1));
    end
    %  ... then we can compute the parametric initial speed from absolute 
    %  initial speed and geometry.
    sdot0 = xdot0 / norm(xs); % why is that? try to find out!
    
    % Informative, not really needed...
    fprintf(1, 'Initial position in 3D space: x=%f m, y=%f m, z=%f m\n', x(1), x(2), x(3));
    
    % Solve ODE (here we do it in advance to the visualisation, but doing
    % it hand in hand with it is trivial)
    tend = 100; % unit: seconds
    delt = 1/15; % unit: seconds (remember: this must be chosen so that the ODE solution is stable!)
    
    % We fix all constant parameters (v,a,b) in advance and pass the
    % resulting functional to the integration methods
    helper_ode = @(t, y)rollercoaster_ode_system(y, v, a, b);
    
    % This is our fourth order Runge-Kutta method (as specified in the
    % document)
    [timepoints ode_solution] = rk4(helper_ode, 0:delt:tend, [s0; sdot0]);
    
    % This is a Euler forward method (just for comparison, it is not stable)
%     [timepoints ode_solution] = euler_forward(helper_ode, 0:delt:tend, [s0; sdot0]);

    % ...but we can also use the more sophisticated ode45 method of MATLAB
    % which implements a Runge-Kutta method of order 5 with a so-called 
    % embedded order 4 method. With the help of these two the time step
    % delt can be chosen *adaptively* so that the error we make is "small
    % enough". The result is interpolated to the required timesteps. See
    % "help ode45" for details.
%     [timepoints ode_solution] = ode45(helper_ode, 0:delt:tend, [s0; sdot0]);

    % =====================================================================
    % Step 4: visualisation of geometry and ODE solution
    spatial_resolution = 0.05;
    frames_per_second = 15;
    
    s = 0:spatial_resolution:1;
    pointset = zeros(dims,length(s));

    for step = 1:length(ode_solution)
        % plot geometry
        clf;
        hold on;
        view(3);
        for i = 1:nodes-1
            for j = 1:length(s)
                pointset(:,j) = bezier_curve(s(j), v(:,i), a(:,i), b(:,i), v(:,i+1));
            end
            plot3(pointset(1,:),pointset(2,:),pointset(3,:)); 
        end
        for j = 1:length(s)
            pointset(:,j) = bezier_curve(s(j), v(:,nodes), a(:,nodes), b(:,nodes), v(:,1));
        end
        plot3(pointset(1,:),pointset(2,:),pointset(3,:));

        % plot spline control points in green
        plot3(a(1,:), a(2,:), a(3,:), 'gx');
        plot3(b(1,:), b(2,:), b(3,:), 'gx');

        % plot nodes in blue
        plot3(v(1,:), v(2,:), v(3,:), 'bx');

        % transform ODE solution to geometry space...
        node = mod(floor(ode_solution(step,1)), nodes) + 1;
        sr = ode_solution(step,1) - floor(ode_solution(step,1));
        if (node ~= nodes)
            [x xs] = bezier_curve(sr, v(:,node), a(:,node), b(:,node), v(:,node+1));
        else
            [x xs] = bezier_curve(sr, v(:,node), a(:,node), b(:,node), v(:,1));
        end
        % ... and plot it as a red circle
        plot3(x(1), x(2), x(3), 'ro');
        
        % print current speed (not needed, only nice)
        fprintf(1, 'Speed: %.1f meters/s\n', norm(xs*ode_solution(step,2)));
        
        % short pause, frames per second (fps)
        pause(1/frames_per_second);
    end
end

% DS_BEZIER_CURVE - small helper function to compute the arc length of the
%   rollercoaster trac.
function [ds] = ds_bezier_curve(s, x1, x2, x3, x4)
    [x, dx] = bezier_curve(s, x1, x2, x3, x4);
    ds = zeros(1, length(s));
    for i = 1:length(s)
        ds(i) = norm(dx(:,i), 2);
    end
end